@staticmethod
def deserialize(bytes):
    """
    Opens a bintensors lazily and returns tensors as asked

    Args:
        data (`bytes`):
            The byte content of a file

    Returns:
        (`List[str, Dict[str, Dict[str, any]]]`):
            The deserialized content is like:
                [("tensor_name", {"shape": [2, 3], "dtype": "float32", "data": b"..." }), (...)]
    """
    pass

@staticmethod
def serialize(tensor_dict, metadata=None):
    """
    Serializes raw data.

    Args:
        tensor_dict (`Dict[str, Dict[Any]]`):
            The tensor dict is like:
                {"tensor_name": {"dtype": "float32", "shape": [2, 3], "data": b"...."}}
        metadata (`Dict[str, str]`, *optional*):
            The optional purely text annotations

    Returns:
        (`bytes`):
            The serialized content.
    """
    pass

@staticmethod
def serialize_file(tensor_dict, filename, metadata=None):
    """
    Serializes raw data into file.

    Args:
        tensor_dict (`Dict[str, Dict[Any]]`):
            Dictorary of tensor data types, and there assoisated layer key
        filename (`str`, or `os.PathLike`):
            The name of the file to write into.
        metadata (`Dict[str, str]`, *optional*):
            The optional purely text annotations

    Returns:
        (`NoneType`):
            On success return `None`.
    """
    pass

class safe_open:
    """
    Opens a bintensors lazily and returns tensors as asked

    Args:
        filename (`str`, or `os.PathLike`):
            The filename to open

        framework (`str`):
            The framework you want you tensors in. Supported values:
            `pt`, `numpy`.

        device (`str`, defaults to `"cpu"`):
            The device on which you want the tensors.
    """

    def __init__(self, filename, framework, device=...):
        pass
    def __enter__(self):
        """
        Start the context manager
        """
        pass
    def __exit__(self, _exc_type, _exc_value, _traceback):
        """
        Exits the context manager
        """
        pass
    def get_slice(self, name):
        """
        Returns a full slice view object

        Args:
            name (`str`):
                The name of the tensor you want

        Returns:
            (`PySafeSlice`):
                A dummy object you can slice into to get a real tensor
        Example:
        ```python
        from bintensors import safe_open

        with safe_open("model.bintensors", framework="pt", device=0) as f:
            tensor_part = f.get_slice("embedding")[:, ::8]

        ```
        """
        pass
    def get_tensor(self, name):
        """
        Returns a full tensor

        Args:
            name (`str`):
                The name of the tensor you want

        Returns:
            (`Tensor`):
                The tensor in the framework you opened the file for.

        Example:
        ```python
        from bintensors import safe_open

        with safe_open("model.bintensors", framework="pt", device=0) as f:
            tensor = f.get_tensor("embedding")

        ```
        """
        pass
    def keys(self):
        """
        Returns the names of the tensors in the file.

        Returns:
            (`List[str]`):
                The name of the tensors contained in that file
        """
        pass
    def metadata(self):
        """
        Return the special non tensor information in the header

        Returns:
            (`Dict[str, str]`):
                The freeform metadata.
        """
        pass

class BintensorError(Exception):
    """
    Custom Python Exception for Safetensor errors.
    """



#!/usr/bin/env python3
import time
import torch
import pyperf
import logging

from typing import Dict, Tuple

from bintensors.torch import serialize, load, _flatten

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)

py_runner = pyperf.Runner()


def create_gpt2(n_layers: int = 20) -> Dict[str, torch.Tensor]:
    tensors = {}
    tensors["wte"] = torch.zeros((50257, 768))
    tensors["wpe"] = torch.zeros((1024, 768))

    for i in range(n_layers):
        tensors[f"h.{i}.ln_1.weight"] = torch.zeros((768,))
        tensors[f"h.{i}.ln_1.bias"] = torch.zeros((768,))
        tensors[f"h.{i}.attn.bias"] = torch.zeros((1, 1, 1024, 1024))
        tensors[f"h.{i}.attn.c_attn.weight"] = torch.zeros((768, 2304))
        tensors[f"h.{i}.attn.c_attn.bias"] = torch.zeros((2304))
        tensors[f"h.{i}.attn.c_proj.weight"] = torch.zeros((768, 768))
        tensors[f"h.{i}.attn.c_proj.bias"] = torch.zeros((768))
        tensors[f"h.{i}.ln_2.weight"] = torch.zeros((768))
        tensors[f"h.{i}.ln_2.bias"] = torch.zeros((768))
        tensors[f"h.{i}.mlp.c_fc.weight"] = torch.zeros((768, 3072))
        tensors[f"h.{i}.mlp.c_fc.bias"] = torch.zeros((3072))
        tensors[f"h.{i}.mlp.c_proj.weight"] = torch.zeros((3072, 768))
        tensors[f"h.{i}.mlp.c_proj.bias"] = torch.zeros((768))

    tensors["ln_f.weight"] = torch.zeros((768))
    tensors["ln_f.bias"] = torch.zeros((768))

    return tensors


def serialize_gpt2() -> Tuple[bytes, float]:
    """Serialize small GPT-2 model and return the buffer and time taken."""
    logger.info(f"Creating GPT-2 model with 20 layers")
    tensor_dict = create_gpt2()

    # Calculate size
    space = _flatten(tensor_dict)
    bytes_used = sum(len(v["data"]) for _, v in space.items())
    logger.info(f"{bytes_used * 1e-6:.2f} MB allocated (not including metadata)")

    t0 = time.perf_counter()
    serialized_buffer = serialize(space)
    serialization_time = time.perf_counter() - t0

    logger.info(f"Serialization took {serialization_time:.4f} seconds")
    logger.info(f"Serialized size: {len(serialized_buffer) * 1e-6:.2f} MB")

    return serialized_buffer, serialization_time


def load_model_bench(buffer: bytes) -> float:
    """Load model from buffer and return the time taken."""
    t0 = time.perf_counter()
    _ = load(buffer)
    loading_time = time.perf_counter() - t0

    logger.info(f"Loading took {loading_time:.4f} seconds")

    return loading_time


def main():
    # Run the "serialize_gpt2" benchmark
    py_runner.bench_func("serialize_gpt2_bench", serialize_gpt2)

    # Run the "load_gpt2" benchmark
    serialized_buffer, _ = serialize_gpt2()
    py_runner.bench_func("load_gpt2_bench", load_model_bench, *[serialized_buffer])


if __name__ == "__main__":
    main()



import pytest

import tempfile
import numpy as np

import mlx.core as mlx


from typing import Dict, Tuple
from bintensors.mlx import load, load_file, save, save_file, safe_open, save_with_checksum


def _compare_mlx_array(lhs: mlx.array, rhs: mlx.array) -> bool:
    return mlx.array_equal(lhs, rhs)


def _ASSERT_SHAPE(value: Tuple[int, ...], expected: Tuple[int, ...]) -> bool:
    assert value == expected, f"assert mismatch of shapes of a tensor, found {value}, expected {expected}"


def create_gpt2_numpy_dict(n_layers: int) -> Dict[str, mlx.array]:
    tensors = {}
    tensors["wte"] = mlx.zeros((50257, 768))
    tensors["wpe"] = mlx.zeros((1024, 768))
    for i in range(n_layers):
        tensors[f"h.{i}.ln_1.weight"] = mlx.zeros((768,))
        tensors[f"h.{i}.ln_1.bias"] = mlx.zeros((768,))
        tensors[f"h.{i}.attn.bias"] = mlx.zeros((1, 1, 1024, 1024))
        tensors[f"h.{i}.attn.c_attn.weight"] = mlx.zeros((768, 2304))
        tensors[f"h.{i}.attn.c_attn.bias"] = mlx.zeros((2304))
        tensors[f"h.{i}.attn.c_proj.weight"] = mlx.zeros((768, 768))
        tensors[f"h.{i}.attn.c_proj.bias"] = mlx.zeros((768))
        tensors[f"h.{i}.ln_2.weight"] = mlx.zeros((768))
        tensors[f"h.{i}.ln_2.bias"] = mlx.zeros((768))
        tensors[f"h.{i}.mlp.c_fc.weight"] = mlx.zeros((768, 3072))
        tensors[f"h.{i}.mlp.c_fc.bias"] = mlx.zeros((3072))
        tensors[f"h.{i}.mlp.c_proj.weight"] = mlx.zeros((3072, 768))
        tensors[f"h.{i}.mlp.c_proj.bias"] = mlx.zeros((768))
    tensors["ln_f.weight"] = mlx.zeros((768))
    tensors["ln_f.bias"] = mlx.zeros((768))
    return tensors

def test_save_and_load_gpt2_tensors_dict_mlx():
    small_gpt2 = create_gpt2_numpy_dict(2)
    buffer = save(small_gpt2)
    model = load(buffer)
    assert all((_compare_mlx_array(small_gpt2[key], model[key]) for key in small_gpt2.keys()))


def test_save_and_load_zero_sized_tensors_mlx():
    _SHAPES_A, _SHAPES_B = (0,), (0, 0, 0)
    tensor_dict = {"ln.weight": mlx.zeros(_SHAPES_A), "ln.bias": mlx.zeros(_SHAPES_B)}

    buffer = save(tensor_dict)
    loaded_tensor_dict = load(buffer)
    _ASSERT_SHAPE(tuple(loaded_tensor_dict["ln.weight"].shape), _SHAPES_A)
    _ASSERT_SHAPE(tuple(loaded_tensor_dict["ln.bias"].shape), _SHAPES_B)


def test_tensor_dtype_roundtrip_mlx():
    tensor_dict = {
        "float32": mlx.zeros((10, 10), dtype=mlx.float32),
        "float64": mlx.zeros((10, 10), dtype=mlx.float32),
        "int32": mlx.zeros((10, 10), dtype=mlx.int32),
        "int64": mlx.zeros((10, 10), dtype=mlx.float32),
    }
    buffer = save(tensor_dict)
    loaded_dict = load(buffer)

    for key, value in tensor_dict.items():
        assert _compare_mlx_array(loaded_dict[key], value)
        assert loaded_dict[key].dtype == value.dtype


def test_save_and_load_large_gpt2_dict_mlx():
    tensor_dict = create_gpt2_numpy_dict(20)
    buffer = save(tensor_dict)
    loaded_dict = load(buffer)
    assert all((_compare_mlx_array(loaded_dict[key], value) for key, value in tensor_dict.items()))


def test_tensor_dict_with_mixed_shapes_mlx():
    tensor_dict = {
        "scalar": mlx.array(5.0),
        "vector": mlx.array([1.0, 2.0, 3.0]),
        "matrix": mlx.array([[1.0, 2.0], [3.0, 4.0]]),
        "tensor3d": mlx.random.normal((2, 3, 4)),
    }

    buffer = save(tensor_dict)
    loaded_dict = load(buffer)

    for key, value in tensor_dict.items():
        assert _compare_mlx_array(loaded_dict[key], value)
        assert loaded_dict[key].shape == value.shape


def test_invalid_tensor_dict_raises_error_mlx():
    invalid_dict = {"valid": mlx.zeros((5, 5)), "invalid": "string_value"}

    with pytest.raises(Exception):
        _ = save(invalid_dict)


def test_save_file_and_load_file_consistency_mlx():
    tensor_dict = create_gpt2_numpy_dict(1)
    filename = ""
    loaded_dict = {}
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        filename = tmp.name
        try:
            save_file(tensor_dict, filename)
            loaded_dict = load_file(filename)
        finally:
            for key, value in tensor_dict.items():
                assert _compare_mlx_array(loaded_dict[key], value)


def test_safe_open_access_and_metadata_mlx():
    tensor_dict = create_gpt2_numpy_dict(1)
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        filename = tmp.name
        # save file into tempfile
        save_file(tensor_dict, filename)

        # contex manager numpy framework allocation on bintensors file
        with safe_open(filename, "numpy") as model:
            assert model.get_tensor("h.0.ln_1.weight") is not None
            assert model.get_tensor("h.0.ln_1.bias") is not None
            assert model.metadata() is None


def test_safe_open_access_with_metadata_mlx():
    tensor_dict = create_gpt2_numpy_dict(1)
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        filename = tmp.name

        save_file(tensor_dict, filename, metadata={"hello": "world"})
        with safe_open(filename, "numpy") as model:
            assert model.get_tensor("h.0.ln_1.weight") is not None
            assert model.get_tensor("h.0.ln_1.bias") is not None
            assert model.metadata()["hello"] == "world"


def test_checksum_two_diffrent_models_mlx():
    model_1 = {"ln.weight": mlx.random.normal((10, 10)), "ln.bias": mlx.random.normal((10,))}
    model_2 = {"ln.weight": mlx.random.normal((10, 10)), "ln.bias": mlx.random.normal((10,))}

    checksum1, _ = save_with_checksum(model_1)
    checksum2, _ = save_with_checksum(model_2)

    assert checksum1 != checksum2, "These checksum are not equivilent"


def test_checksum_two_same_models_mlx():
    model_1 = {"ln.weight": mlx.zeros((2, 2)), "ln.bias": mlx.zeros((10))}
    model_2 = {"ln.weight": mlx.zeros((2, 2)), "ln.bias": mlx.zeros((10))}

    for _ in range(1000):
        checksum1, _ = save_with_checksum(model_1)
        checksum2, _ = save_with_checksum(model_2)
        assert checksum1 == checksum2, "These checksum are equivilent"



import os
import sys
import hashlib
from _hashlib import HASH
from typing import Dict, Optional, Union, Tuple, Callable

try:
    import numpy as np
except ImportError:
    raise ImportError(
        "Could not find the 'numpy' module. To use this part of the package, please install numpy: `pip install numpy`."
    )


from bintensors import deserialize, safe_open, serialize, serialize_file

__all__ = ["save", "save_file", "load", "load_file", "save_with_checksum"]


def _tobytes(tensor: np.ndarray) -> bytes:
    """
    Converts a `np.ndarray` into a raw little-endian byte representation.

    Args:
        tensor (`np.ndarray`):
            A dense and contiguous NumPy array.

    Returns:
        `bytes`: A byte representation of the array's data in little-endian order.
    """
    if not _is_little_endian(tensor):
        tensor = tensor.byteswap(inplace=False)
    return tensor.tobytes()


def save(tensor_dict: Dict[str, np.ndarray], metadata: Optional[Dict[str, str]] = None) -> bytes:
    """
    Saves a dictionary of tensors into raw bytes in bintensors format.

    Args:
        tensor_dict (`Dict[str, np.ndarray]`):
            The incoming tensors. Tensors need to be contiguous and dense.
        metadata (`Dict[str, str]`, *optional*, defaults to `None`):
            Optional text only metadata you might want to save in your header.
            For instance it can be useful to specify more about the underlying
            tensors. This is purely informative and does not affect tensor loading.

    Returns:
        `bytes`: The raw bytes representing the format

    Example:

    ```python
    from bintensors.numpy import save
    import numpy as np

    tensors = {"embedding": np.zeros((512, 1024)), "attention": np.zeros((256, 256))}
    byte_data = save(tensors)
    ```
    """
    flattened = {k: {"dtype": v.dtype.name, "shape": v.shape, "data": _tobytes(v)} for k, v in tensor_dict.items()}
    serialized = serialize(flattened, metadata=metadata)
    result = bytes(serialized)
    return result


def save_file(
    tensor_dict: Dict[str, np.ndarray], filename: Union[str, os.PathLike], metadata: Optional[Dict[str, str]] = None
) -> None:
    """
    Saves a dictionary of tensors into raw bytes in bintensors format.

    Args:
        tensor_dict (`Dict[str, np.ndarray]`):
            The incoming tensors. Tensors need to be contiguous and dense.
        filename (`str`, or `os.PathLike`)):
            The filename we're saving into.
        metadata (`Dict[str, str]`, *optional*, defaults to `None`):
            Optional text only metadata you might want to save in your header.
            For instance it can be useful to specify more about the underlying
            tensors. This is purely informative and does not affect tensor loading.

    Returns:
        `None`

    Example:

    ```python
    from bintensors.numpy import save_file
    import numpy as np

    tensors = {"embedding": np.zeros((512, 1024)), "attention": np.zeros((256, 256))}
    save_file(tensors, "model.bintensors")
    ```
    """
    flattened = {k: {"dtype": v.dtype.name, "shape": v.shape, "data": _tobytes(v)} for k, v in tensor_dict.items()}
    serialize_file(filename, flattened, metadata=metadata)


def save_with_checksum(
    tensor_dict: Dict[str, np.ndarray],
    metadata: Optional[Dict[str, str]] = None,
    hasher: Callable[[bytes], HASH] = hashlib.sha1,
) -> Tuple[bytes, bytes]:
    """
    Saves a dictionary of tensors into raw bytes in bintensors format.

    Args:
        tensors (`Dict[str, torch.Tensor]`):
            The incoming tensors. Tensors need to be contiguous and dense.
        metadata (`Dict[str, str]`, *optional*, defaults to `None`):
            Optional text only metadata you might want to save in your header.
            For instance it can be useful to specify more about the underlying
            tensors. This is purely informative and does not affect tensor loading.
        hasher (`Callable[[bytes], HASH]`):
            A hash is an object used to calculate a checksum of a string of information.

    Returns:
        `bytes`: The raw bytes representing the format

    Example:

    ```python
    from bintensors.torch import save_with_checksum
    import torch

    tensors = {"embedding": torch.zeros((512, 1024)), "attention": torch.zeros((256, 256))}
    checksum, byte_data = save_with_checksum(tensors)
    ```
    """
    buffer = save(tensor_dict, metadata=metadata)
    buffer = bytes(buffer)
    result = hasher(buffer).digest(), buffer
    return result


def load(data: bytes) -> Dict[str, np.ndarray]:
    """
    Loads a bintensors file into numpy format from pure bytes.

    Args:
        data (`bytes`):
            The content of a bintensors file

    Returns:
        `Dict[str, np.ndarray]`: dictionary that contains name as key, value as `np.ndarray` on cpu

    Example:

    ```python
    from bintensors.numpy import load

    file_path = "./my_folder/bert.bintensors"
    with open(file_path, "rb") as f:
        data = f.read()

    loaded = load(data)
    ```
    """
    flat = deserialize(data)
    return _view2np(flat)


def load_file(filename: Union[str, os.PathLike]) -> Dict[str, np.ndarray]:
    """
    Loads a bintensors file into numpy format.

    Args:
        filename (`str`, or `os.PathLike`)):
            The name of the file which contains the tensors

    Returns:
        `Dict[str, np.ndarray]`: dictionary that contains name as key, value as `np.ndarray`

    Example:

    ```python
    from bintensors.numpy import load_file

    file_path = "./my_folder/bert.bintensors"
    loaded = load_file(file_path)
    ```
    """
    result = {}
    with safe_open(filename, framework="np") as f:
        for k in f.offset_keys():
            result[k] = f.get_tensor(k)
    return result


# np.float8 formats require 2.1; we do not support these dtypes on earlier versions
_TYPES = {
    "F64": np.float64,
    "F32": np.float32,
    "F16": np.float16,
    "I64": np.int64,
    "U64": np.uint64,
    "I32": np.int32,
    "U32": np.uint32,
    "I16": np.int16,
    "U16": np.uint16,
    "I8": np.int8,
    "U8": np.uint8,
    "BOOL": bool,
}


def _getdtype(dtype_str: str) -> Optional[np.dtype]:
    """
    Map bintensors string to numpy data type.

    Args:
        dtype_str (`str`):
            string repersentation of the `np.dtype`.

    Returns:
        `Optional[np.dtype]`: data type repersentation of the tensors object, if such dtype does not exist retrun None.
    """
    return _TYPES.get(dtype_str, None)


def _view2np(safeview) -> Dict[str, np.ndarray]:
    """
    Convert a view to a numpy array object

    Args:
        safeview (`Dict[str, Union[bytes, str, Tuple[int,...]]]`)
            object view of the tensors within the bintensor file format

    Returns:
        `Dict[str, np.ndarray]`: dictionary of layer, and numpy array objects.
    """
    result = {}
    for k, v in safeview:
        dtype = _getdtype(v["dtype"])
        arr = np.frombuffer(v["data"], dtype=dtype).reshape(v["shape"])
        result[k] = arr
    return result


def _is_little_endian(tensor: np.ndarray) -> bool:
    """
    Check the byte order is a little-endian of the tensors

    Args:
        tensor (`np.ndarray`):
            numpy data array tensor object.

    Returns:
        `bool`: True if the tensor ``object.dtype.byteorder`` is laid out little endian.
    """
    byteorder = tensor.dtype.byteorder
    if byteorder == "=":
        if sys.byteorder == "little":
            return True
        else:
            return False
    elif byteorder == "|":
        return True
    elif byteorder == "<":
        return True
    elif byteorder == ">":
        return False
    raise ValueError(f"Unexpected byte order {byteorder}")



import pytest
import struct

import torch
from typing import List, Dict, Tuple
from itertools import chain

from bintensors import BintensorError
from bintensors.torch import save, load

_DTYPE = {
    "BOL": 0,
    "U8": 1,
    "I8": 2,
    "F8_E5M2": 3,
    "F8_E4M3": 4,
    "I16": 5,
    "U16": 6,
    "F16": 7,
    "BF16": 8,
    "I32": 9,
    "U32": 10,
    "F32": 11,
    "F64": 12,
    "I64": 13,
    "F64": 14,
}


def encode_unsigned_variant_encoding(number: int) -> bytes:
    """Encodes an unsigned integer into a variable-length format."""
    if number > 0xFFFFFFFF:
        return b"\xfd" + number.to_bytes(8, "little")
    elif number > 0xFFFF:
        return b"\xfc" + number.to_bytes(4, "little")
    elif number > 0xFA:
        return b"\xfb" + number.to_bytes(2, "little")
    else:
        return number.to_bytes(1, "little")


def encode_header(id: str, dtype: str, shape: Tuple[int, ...], offset: Tuple[int, int]) -> bytes:
    """Encodes the struct TensorInfo into byte buffer with string ID prefix."""
    if dtype not in _DTYPE:
        raise ValueError(f"Unsupported dtype: {dtype}")

    encoded_id = encode_unsigned_variant_encoding(len(id)) + id.encode("utf-8")

    # Compose numeric fields
    numeric_layout = chain([_DTYPE[dtype], len(shape)], shape, offset)

    encoded_tensor_info = b"".join(encode_unsigned_variant_encoding(x) for x in numeric_layout)

    return encoded_id + encoded_tensor_info


def test_empty_file():
    "bintensors allows empty dictonary"
    tensor_dict = {}
    buffer = save(tensor_dict)
    # decouple first 8 bytes part of the buffer unsinged long long
    header_size = struct.unpack("<Q", buffer[0:8])[0]
    # header size + metadata + empty tensors
    MAX_FILE_SIZE = 8 + header_size
    assert header_size == 8, "expected packed buffer shoudl be unsinged interger 8."
    assert buffer[8:] == b"\x00\x00      ", "expected empty metadata fields."
    assert MAX_FILE_SIZE == len(buffer), "These should  be equal"


def test_man_cmp():
    size = 2
    shape = (2, 2)
    tensor_chunk_length = shape[0] * shape[1] * 4  # Size of a tensor buffer

    length = encode_unsigned_variant_encoding(size)

    # Create tensor info buffer
    tensor_info_buffer = b"".join(
        encode_header(
            f"weight_{i}",
            "F32",
            shape,
            (i * tensor_chunk_length, i * tensor_chunk_length + tensor_chunk_length),
        )
        for i in range(size)
    )
    layout = length + tensor_info_buffer
    layout = b"\0" + layout
    layout += b" " * (((8 - len(layout)) % 8) % 8)
    n = len(layout)
    n_header = n.to_bytes(8, "little")

    expected = n_header + layout + (b"\0" * tensor_chunk_length * size)

    tensor_dict = {"weight_0": torch.zeros(shape), "weight_1": torch.zeros(shape)}

    buffer = save(tensor_dict)
    # we need to check both since there is no order in the hashmap
    assert buffer == expected, f"got {buffer}, and expected {expected}"


def test_missmatch_length_of_metadata_large():
    size = 2
    shape = (2, 2)
    tensor_chunk_length = shape[0] * shape[1] * 4  # Size of a tensor buffer

    length = encode_unsigned_variant_encoding(size * 1000)

    # Create tensor info buffer
    tensor_info_buffer = b"".join(
        encode_header(
            f"weight_{i}",
            "F32",
            shape,
            (i * tensor_chunk_length, i * tensor_chunk_length + tensor_chunk_length),
        )
        for i in range(size)
    )
    layout = length + tensor_info_buffer
    layout = b"\0" + layout
    layout += b" " * (((8 - len(layout)) % 8) % 8)
    n = len(layout)
    n_header = n.to_bytes(8, "little")

    # layout together
    buffer = n_header + layout + b"\0" * (tensor_chunk_length * size)

    with pytest.raises(BintensorError):
        # this is not a valid since the metadata
        # size doe not match as it too big
        _ = load(buffer)


def test_missmatch_length_of_metadata_small():
    size = 2
    shape = (2, 2)
    tensor_chunk_length = shape[0] * shape[1] * 4  # Size of a tensor buffer

    length = encode_unsigned_variant_encoding(size - 1)

    # Create tensor info buffer
    tensor_info_buffer = b"".join(
        encode_header(
            f"weight_{i}",
            "F32",
            shape,
            (i * tensor_chunk_length, i * tensor_chunk_length + tensor_chunk_length),
        )
        for i in range(size)
    )
    layout = length + tensor_info_buffer
    layout = b"\0" + layout
    layout += b" " * (((8 - len(layout)) % 8) % 8)
    n = len(layout)
    n_header = n.to_bytes(8, "little")

    # layout together
    buffer = n_header + layout + b"\0" * (tensor_chunk_length * size)

    with pytest.raises(BintensorError):
        # this is not a valid since the metadata
        # size doe not match as it too big
        _ = load(buffer)



import os
import hashlib
from _hashlib import HASH
from typing import Dict, Optional, Union, Tuple, Callable

import numpy as np

import mlx.core as mx
from bintensors import numpy, safe_open

__all__ = ["save", "save_file", "load", "load_file", "save_with_checksum"]

def save(tensors: Dict[str, mx.array], metadata: Optional[Dict[str, str]] = None) -> bytes:
    """
    Saves a dictionary of tensors into raw bytes in bintensors format.

    Args:
        tensors (`Dict[str, mx.array]`):
            The incoming tensors. Tensors need to be contiguous and dense.
        metadata (`Dict[str, str]`, *optional*, defaults to `None`):
            Optional text only metadata you might want to save in your header.
            For instance it can be useful to specify more about the underlying
            tensors. This is purely informative and does not affect tensor loading.

    Returns:
        `bytes`: The raw bytes representing the format

    Example:

    ```python
    from bintensors.mlx import save
    import mlx.core as mx

    tensors = {"embedding": mx.zeros((512, 1024)), "attention": mx.zeros((256, 256))}
    byte_data = save(tensors)
    ```
    """
    np_tensors = _mx2np(tensors)
    return numpy.save(np_tensors, metadata=metadata)


def save_file(
    tensors: Dict[str, mx.array],
    filename: Union[str, os.PathLike],
    metadata: Optional[Dict[str, str]] = None,
) -> None:
    """
    Saves a dictionary of tensors into raw bytes in bintensors format.

    Args:
        tensors (`Dict[str, mx.array]`):
            The incoming tensors. Tensors need to be contiguous and dense.
        filename (`str`, or `os.PathLike`)):
            The filename we're saving into.
        metadata (`Dict[str, str]`, *optional*, defaults to `None`):
            Optional text only metadata you might want to save in your header.
            For instance it can be useful to specify more about the underlying
            tensors. This is purely informative and does not affect tensor loading.

    Returns:
        `None`

    Example:

    ```python
    from bintensors.mlx import save_file
    import mlx.core as mx

    tensors = {"embedding": mx.zeros((512, 1024)), "attention": mx.zeros((256, 256))}
    save_file(tensors, "model.bt")
    ```
    """
    np_tensors = _mx2np(tensors)
    return numpy.save_file(np_tensors, filename, metadata=metadata)


def save_with_checksum(
    tensor_dict: Dict[str, np.ndarray],
    metadata: Optional[Dict[str, str]] = None,
    hasher: Callable[[bytes], HASH] = hashlib.sha1,
) -> Tuple[bytes, bytes]:
    """
    Saves a dictionary of tensors into raw bytes in bintensors format.

    Args:
        tensors (`Dict[str, np.ndarray]`):
            The incoming tensors. Tensors need to be contiguous and dense.
        metadata (`Dict[str, str]`, *optional*, defaults to `None`):
            Optional text only metadata you might want to save in your header.
            For instance it can be useful to specify more about the underlying
            tensors. This is purely informative and does not affect tensor loading.
        hasher (`Callable[[bytes], HASH]`):
            A hash is an object used to calculate a checksum of a string of information.


    Returns:
        `bytes`: The raw bytes representing the format

    Example:

    ```python
    from bintensors.mlx import save_with_checksum
    import mlx.core as mx

    tensors = {"embedding": mx.zeros((512, 1024)), "attention": mx.zeros((256, 256))}
    checksum, byte_data = save_with_checksum(tensors)
    ```
    """
    np_tensors = _mx2np(tensor_dict)
    return numpy.save_with_checksum(np_tensors, metadata, hasher)


def load(data: bytes) -> Dict[str, mx.array]:
    """
    Loads a bintensors file into MLX format from pure bytes.

    Args:
        data (`bytes`):
            The content of a bintensors file

    Returns:
        `Dict[str, mx.array]`: dictionary that contains name as key, value as `mx.array`

    Example:

    ```python
    from bintensors.mlx import load

    file_path = "./my_folder/bert.bt"
    with open(file_path, "rb") as f:
        data = f.read()

    loaded = load(data)
    ```
    """
    flat = numpy.load(data)
    return _np2mx(flat)


def load_file(filename: Union[str, os.PathLike]) -> Dict[str, mx.array]:
    """
    Loads a bintensors file into MLX format.

    Args:
        filename (`str`, or `os.PathLike`)):
            The name of the file which contains the tensors

    Returns:
        `Dict[str, mx.array]`: dictionary that contains name as key, value as `mx.array`

    Example:

    ```python
    from bintensors.flax import load_file

    file_path = "./my_folder/bert.bt"
    loaded = load_file(file_path)
    ```
    """
    result = {}
    with safe_open(filename, framework="mlx") as f:
        for k in f.offset_keys():
            result[k] = f.get_tensor(k)
    return result


def _np2mx(numpy_dict: Dict[str, np.ndarray]) -> Dict[str, mx.array]:
    """
    Preform conversion from numpy storage backend to mlx storage backend.

    Args:
        tensors (`Dict[str,  np.ndarray]`):
            The incoming tensors. Tensors need to be contiguous and dense.

    Returns:
          `Dict[str, mx.array]`: dictionary that contains name as key, value as `mx.array`
    """
    for k, v in numpy_dict.items():
        numpy_dict[k] = mx.array(v)
    return numpy_dict


def _mx2np(mx_dict: Dict[str, mx.array]) -> Dict[str, np.ndarray]:
    """
    Preform conversion from mlx storage backend to numpy storage backend.

    Args:
        tensors (`Dict[str, mx.array]`):
            The incoming tensors. Tensors need to be contiguous and dense.
    Returns:
          `Dict[str, np.ndarray]`: dictionary that contains name as key, value as `np.ndarray`
    """
    new_dict = {}
    for k, v in mx_dict.items():
        new_dict[k] = np.asarray(v)
    return new_dict



import pytest

import os
import tempfile
import torch

from typing import Dict, Tuple
from bintensors.torch import load, save, save_file, load_file, safe_open, save_with_checksum


def _compare_torch_tensors(lhs: torch.Tensor, rhs: torch.Tensor) -> bool:
    return torch.equal(lhs, rhs)


def _ASSERT_SHAPE(value: Tuple[int, ...], expected: Tuple[int, ...]) -> bool:
    assert value == expected, f"assert mismatch of shapes of a tensor, found {value}, expected {expected}"


def create_gpt2_tensors_dict(n_layers: int) -> Dict[str, torch.Tensor]:
    tensors = {}
    tensors["wte"] = torch.zeros((50257, 768))
    tensors["wpe"] = torch.zeros((1024, 768))
    for i in range(n_layers):
        tensors[f"h.{i}.ln_1.weight"] = torch.zeros((768,))
        tensors[f"h.{i}.ln_1.bias"] = torch.zeros((768,))
        tensors[f"h.{i}.attn.bias"] = torch.zeros((1, 1, 1024, 1024))
        tensors[f"h.{i}.attn.c_attn.weight"] = torch.zeros((768, 2304))
        tensors[f"h.{i}.attn.c_attn.bias"] = torch.zeros((2304))
        tensors[f"h.{i}.attn.c_proj.weight"] = torch.zeros((768, 768))
        tensors[f"h.{i}.attn.c_proj.bias"] = torch.zeros((768))
        tensors[f"h.{i}.ln_2.weight"] = torch.zeros((768))
        tensors[f"h.{i}.ln_2.bias"] = torch.zeros((768))
        tensors[f"h.{i}.mlp.c_fc.weight"] = torch.zeros((768, 3072))
        tensors[f"h.{i}.mlp.c_fc.bias"] = torch.zeros((3072))
        tensors[f"h.{i}.mlp.c_proj.weight"] = torch.zeros((3072, 768))
        tensors[f"h.{i}.mlp.c_proj.bias"] = torch.zeros((768))
    tensors["ln_f.weight"] = torch.zeros((768))
    tensors["ln_f.bias"] = torch.zeros((768))
    return tensors


class ToyRegressionModel(torch.nn.Module):
    def __init__(self, n_in: int, n_out: int):
        super().__init__()
        self.ln_1 = torch.nn.Linear(n_in, n_out)
        self.output = torch.nn.Linear(n_in, 1)

    def forward(self, x: torch.Tensor):
        x = self.ln_1(x)
        x = torch.nn.functional.sigmoid(x)
        x = self.output(x)
        return x


def test_pt_save_and_load_gpt2_tensors_dict():
    small_gpt2 = create_gpt2_tensors_dict(2)
    buffer = save(small_gpt2)
    model = load(buffer)
    all((_compare_torch_tensors(small_gpt2[key], model[key]) for key in small_gpt2.keys()))


def test_pt_save_and_load_zero_sized_tensors():
    _SHAPES_A, _SHAPES_B = (0,), (0, 0, 0)
    tensor_dict = {"ln.weight": torch.zeros(_SHAPES_A), "ln.bias": torch.zeros(_SHAPES_B)}

    buffer = save(tensor_dict)
    loaded_tensor_dict = load(buffer)
    _ASSERT_SHAPE(tuple(loaded_tensor_dict["ln.weight"].shape), _SHAPES_A)
    _ASSERT_SHAPE(tuple(loaded_tensor_dict["ln.bias"].shape), _SHAPES_B)


def test_pt_tensor_dtype_roundtrip():
    tensor_dict = {
        "float32": torch.zeros((10, 10), dtype=torch.float32),
        "float64": torch.zeros((10, 10), dtype=torch.float64),
        "int32": torch.zeros((10, 10), dtype=torch.int32),
        "int64": torch.zeros((10, 10), dtype=torch.int64),
    }
    buffer = save(tensor_dict)
    loaded_dict = load(buffer)

    for key, value in tensor_dict.items():
        assert _compare_torch_tensors(loaded_dict[key], value)
        assert loaded_dict[key].dtype == value.dtype


def test_nn_module_toy_model():
    model = ToyRegressionModel(10, 10)

    tensor_dict = model.state_dict()
    buffer = save(tensor_dict)
    loaded_dict = load(buffer)

    assert all((_compare_torch_tensors(loaded_dict[key], value) for (key, value) in tensor_dict.items()))

    model.load_state_dict(state_dict=loaded_dict, strict=True)


def test_pt_invalid_tensor_dict_raises_error():
    invalid_dict = {"valid": torch.zeros((5, 5)), "invalid": "string_value"}

    with pytest.raises(Exception):
        _ = save(invalid_dict)


def test_pt_save_file_and_load_file_consistency():
    tensor_dict = create_gpt2_tensors_dict(1)
    filename = ""
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        filename = tmp.name
        save_file(tensor_dict, filename)
        loaded_dict = load_file(filename)

        for key, value in tensor_dict.items():
            assert _compare_torch_tensors(loaded_dict[key], value)


def test_pt_safe_open_access_and_metadata():
    tensor_dict = create_gpt2_tensors_dict(1)
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        filename = tmp.name
        save_file(tensor_dict, filename)
        with safe_open(filename, "pt") as model:
            assert model.get_tensor("h.0.ln_1.weight") is not None
            assert model.get_tensor("h.0.ln_1.bias") is not None
            assert model.metadata() is None


def test_checksum_two_diffrent_models():
    model_1 = {"ln.weight": torch.rand((10, 10)), "ln.bias": torch.rand((10))}
    model_2 = {"ln.weight": torch.rand((10, 10)), "ln.bias": torch.rand((10))}

    checksum1, _ = save_with_checksum(model_1)
    checksum2, _ = save_with_checksum(model_2)

    assert checksum1 != checksum2, "These checksum are not equivilent"


def test_checksum_two_same_models():
    model_1 = {"ln.weight": torch.zeros((2, 2)), "ln.bias": torch.zeros((10))}
    model_2 = {"ln.weight": torch.zeros((2, 2)), "ln.bias": torch.zeros((10))}

    for _ in range(1000):
        checksum1, _ = save_with_checksum(model_1)
        checksum2, _ = save_with_checksum(model_2)
        assert checksum1 == checksum2, "These checksum are equivilent"
